{#############################################################################}
{#                                                                           #}
{#  WinQUANT v2 - source code                                                #}
{#  Delphi version                                                           #}
{#  File: Bookmark.pas v2.0                                                  #}
{#                                                                           #}
{#  (C) 2001 Petr Strnsk                                                   #}
{#  All rights reserved                                                      #}
{#                                                                           #}
{#  Web:     http://www.fi.muni.cz/~xstransk/winquant                        #}
{#  E-mail:  xstransk@fi.muni.cz                                             #}
{#           aargh@email.cz                                                  #}
{#                                                                           #}
{#############################################################################}
{#                                                                           #}
{#  This file is distributed as freeware and without warranties of any kind. #}
{#  You can use it in your own applications at your own risk.                #}
{#  See help file for more information.                                      #}
{#                                                                           #}
{#############################################################################}

unit Bookmark;

interface

uses Document, Classes, SysUtils;

const SaveFileVersion = '1.3';                      {do not change length of these strings}
      SaveFileHeader = 'WinQUANT Bookmark File ';

type TBookmark = record
       DocItem: TDocument;
     end;

     TPBookmark = ^TBookmark;

{*** Class TBookmarks is used to store and work with bookmarks ***}
{***                                                           ***}
{*** STRUCTURE:                                                ***}
{***                       TBookmarks                          ***}
{***                            |                              ***}
{***       -------------------------------------  ........     ***}
{***       |                    |                              ***}
{***   TBookmark            TBookmark             ........     ***}
{***       |                    |                              ***}
{***   TDocument            TDocument             ........     ***}


     TBookmarks = class
       public
         Filename: string[255];                {file assigned with this structure}
         UnsavedChanges: boolean;              {are there any changes, which were not saved to file?}
         constructor Create;
         destructor Destroy; override;
         function Count:integer;               {number of stored bookmarks}
         procedure Copy(Source: TBookmarks);   {copy all bookmarks from Source; current bookmarks will be erased}
         procedure Clear;                      {remove all bookmarks and set default values}
         procedure ClearList;                  {remove all bookmarks}
         procedure Sort;                       {sort bookmarks alphabeticaly}
         procedure AddMark(Item: TDocument; Name: string = '');    {add bookmark Item and set its name}
         procedure DeleteMark(Index: integer);                     {remove bookmark}
         function GetMark(Index: integer):TDocument;               {get bookmark}
         function GetName(Index: integer):string;                  {get name of bookmark}
         function FindName(Name: string): integer;                 {find bookmark with specified name}
         procedure CopyMark(Index: integer; Item: TDocument);      {rewrite bookmark with Item}
         procedure MoveMark(FromIndex, ToIndex: integer);          {move bookmark from FromIndex to ToIndex}
         procedure RenameMark(Index: integer; NewName: string);    {rename bookmark}
         procedure SaveToFile;                 {save bookmarks in bookmark file Filename}
         procedure LoadFromFile;               {load bookmarks from file Filename}
       private
         MarkList: TList;                      {list of bookmarks in the file}
     end;

{     EBookmarkError = class(Exception) end;}

implementation

constructor TBookmarks.Create;
begin
  MarkList := TList.Create;
  FileName := '';
  UnsavedChanges := false;
end;

destructor TBookmarks.Destroy;
begin
  ClearList;
  MarkList.Free;
end;

function TBookmarks.Count: integer;
begin
  count := MarkList.Count;
end;

procedure TBookmarks.Clear;
begin
  ClearList;
  FileName := '';
  UnsavedChanges := false;
end;

procedure TBookmarks.Copy(Source: TBookmarks);
var loop: integer;
begin
  FileName := Source.FileName;
  UnsavedChanges := true;
  ClearList;
  for loop:=0 to Source.MarkList.Count-1 do            {clone all bookmarks}
       AddMark(TPDocument(Source.MarkList.Items[loop])^);
end;

procedure TBookmarks.ClearList;
var loop: integer;
begin
  for loop:= 0 to MarkList.Count-1 do begin
     TPBookmark(MarkList.Items[loop])^.DocItem.Destroy; {destroy document in bookmark}
     Dispose(TPBookmark(MarkList.Items[loop]));         {remove bookmark}
  end;
  MarkList.Clear;
end;

  function Compare(Item1, Item2: Pointer): Integer;     {compare function}
  var Name1, Name2: string;
  begin
    Name1 := TPDocument(Item1)^.Name;
    Name2 := TPDocument(Item2)^.Name;
    if Name1 < Name2 then Result := -1
    else if Name1 > Name2 then Result := 1
    else Result := 0;
  end;

procedure TBookmarks.Sort;
begin
  MarkList.Sort(Compare);
end;

procedure TBookmarks.AddMark(Item: TDocument; Name: string);
var InitDoc: TDocument;
    PBookmark: TPBookmark;
    Index: integer;
begin
  if Name = '' then Name := Item.Name;
  Name := Trim(Name);
  if Name = '' then Exit;
  UnsavedChanges := true;
  Index := FindName(Name);
  if Index <> -1 then begin        {does the bookmark already exist in the list?}
    CopyMark(Index, Item);         {no?}
    GetMark(Index).Name := Name;
  end else begin                   {yes?}
    New(PBookmark);
    InitDoc := TDocument.Create;
    InitDoc.Copy(Item);
    InitDoc.Name := Name;
    PBookmark.DocItem := InitDoc;
    MarkList.Add(PBookmark);
  end;
end;

procedure TBookmarks.DeleteMark(Index: integer);      {remove bookmark}
begin
  try
    TPBookmark(MarkList.Items[Index])^.DocItem.Destroy;
    Dispose(TPBookmark(MarkList.Items[Index]));
    MarkList.Delete(Index);
  except
    {ignore all exceptions}
  end;
end;

function TBookmarks.GetMark(Index: integer):TDocument;
begin
  getmark := TPBookmark(MarkList.Items[index])^.DocItem;
end;

function TBookmarks.GetName(Index: integer):string;
begin
  getname := TPBookmark(MarkList.Items[index])^.DocItem.Name;
end;

function TBookmarks.FindName(Name: string): integer;
begin
  Name := Trim(Name);
  for Result := 0 to MarkList.Count-1 do
    if GetName(Result) = Name then Exit;
  Result := -1;
end;

procedure TBookmarks.CopyMark(Index: integer; Item: TDocument);
begin
  GetMark(Index).Copy(Item);
  UnsavedChanges := true;
end;

procedure TBookmarks.MoveMark(FromIndex, ToIndex: integer);
begin
  try
    MarkList.Move(FromIndex, ToIndex);
  except
  end;
end;

procedure TBookmarks.RenameMark(Index: integer; NewName: string);
begin
  if FindName(NewName) <> Index then begin      {don't rename bookmark to itself}
    AddMark(GetMark(Index), NewName);
    DeleteMark(Index);
    MoveMark(Count-1, Index);
  end;
end;

procedure TBookmarks.SaveToFile;
var F: TFileStream;
    loop: integer;
begin
  try
    F := TFileStream.Create(FileName, fmCreate or fmShareExclusive);
    F.WriteBuffer(SaveFileHeader, Length(SaveFileHeader));
    F.WriteBuffer(SaveFileVersion, Length(SaveFileVersion));
    F.WriteBuffer(MarkList.Count, SizeOf(MarkList.Count));
    for loop := 0 to MarkList.Count-1 do Getmark(loop).SaveToStream(F);
    F.Free;
    UnsavedChanges := false;
  except
    on EWriteError do ;
  end;
end;

procedure TBookmarks.LoadFromFile;
var loop, NumOfMarks: integer;
    InitDoc: TDocument;
    F: TFileStream;
//    Version: string;
begin
  ClearList;
  InitDoc := TDocument.Create;
  try
    F := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
    F.Seek(Length(SaveFileHeader), soFromCurrent);
    F.Seek(Length(SaveFileVersion), soFromCurrent);
{    F.ReadBuffer(Version, Length(SaveFileVersion));
    if Version <> SaveFileVersion then
      raise EBookmarkError.Create('ERROR: Incompatible file version');}
    F.ReadBuffer(NumOfMarks, SizeOf(NumOfMarks));
    for loop := 0 to NumOfMarks-1 do begin
      InitDoc.LoadFromStream(F);
      AddMark(InitDoc);
    end;
    F.Free;
    UnsavedChanges := false;
  except
    on EReadError do ;
  end;
end;

end.
