{#############################################################################}
{#                                                                           #}
{#  WinQUANT v2 - source code                                                #}
{#  Delphi version                                                           #}
{#  File: Document.pas v2.0                                                  #}
{#                                                                           #}
{#  (C) 2001 Petr Strnsk                                                   #}
{#  All rights reserved                                                      #}
{#                                                                           #}
{#  Web:     http://www.fi.muni.cz/~xstransk/winquant                        #}
{#  E-mail:  xstransk@fi.muni.cz                                             #}
{#           aargh@email.cz                                                  #}
{#                                                                           #}
{#############################################################################}
{#                                                                           #}
{#  This file is distributed as freeware and without warranties of any kind. #}
{#  You can use it in your own applications at your own risk.                #}
{#  See help file for more information.                                      #}
{#                                                                           #}
{#############################################################################}

unit Document;

interface

uses Graphics, Classes, Windows {because of GetRValue};

const MaxFunctionLength = 255;  {maximum length of function string}
      MaxNameLength = 100;      {maximum length of function name}
type
    TColorEffect = (ceSolid, ceRandom,             {types of effects}
                    cePattern, cePatternInvert, ceLeft2Pattern, cePattern2Right,
                    ceLevelling, ceLeft2Levelling, ceLevelling2Right,
                    ceRandomFill);
    TIPStyle = (isLinear, isSquare, isSqRoot);     {types of interpolation}

    TColorInterval = record
      FromPer, ToPer: single;             {interval from/to (%)}
      LeftColor, RightColor: TColor;      {interval colors}
      Effect: TColorEffect;               {how to fill interval area using the colors}
      Pattern: integer;                   {specify pattern (if necessary}
      Invert: boolean;                    {pattern or levelling inverted ?}
      IPStyle: TIPStyle;                  {type of interpolation}
      LevORRand: single;                  {levelling width}
    end;

    TPColorInterval = ^TColorInterval;
    TPDocument = ^TDocument;

    TDocument = class
    public
      Name: string [MaxNameLength];     {document name}
      Func: string [MaxFunctionLength]; {function text}
      XF, XT, YF, YT: single;           {interval bounds}
      RangeF, RangeT: single;           {value range}
      UndefColor: TColor;               {color for undefined values}
    {* GENERAL *}
      constructor Create;
      destructor Destroy; override;
      procedure ClearList;                    {clear the color interval list}
      procedure Copy(Source: TDocument);      {clear and copy from source}
      procedure SaveToStream(Stream: TStream);       {write document to selected stream}
      procedure LoadFromStream(Stream: TStream);     {load document from selected stream}
    {* RESOLUTION *}
      procedure GetRes(out xr, yr: integer);  {return document resolution}
      function GetXRes:integer;               {return only horizontal resolution}
      function GetYRes:integer;               {return only vertical resolution}
      procedure SetRes(xr, yr: integer);      {set document resolution}
    {* COLOR INTERVALS *}
      function IntCount: integer;             {return number of intervals}
      function GetInt(index: integer):TPColorInterval;
                                              {return chosen intervali - index=1..n }
      procedure AddInt(Index: integer; FromPer, ToPer: single;
                       LeftColor, RightColor: TColor;
                       Effect: TColorEffect;
                       Pattern: integer; Invert: boolean;
                       IPStyle: TIPStyle; LevORRand: single); overload;
                                              {add interval}
      procedure AddInt(Index: integer; Source: TColorInterval); overload;
                                              {copy interval}
      procedure DelInt(Index: integer);       {remove interval}
      procedure DelAndResizeInt(Index: integer);
                                              {delete interval and resize the neighbours}
      procedure SplitInt(Index: integer);     {split interval}
      procedure ChangeFromInt(Index: integer; Newrange: single);
                                              {change interval beginning}
      procedure ChangeToInt(Index: integer; Newrange: single);
                                              {change interval end}
      procedure EqualizeIntLengths;           {set equal interval lengths}
      procedure CopyIntColoring(Index: integer; Source: TColorInterval);
                                              {copy color data from source}
      function GetColorAtInt(PercentPos: single; Int: TColorInterval; PatPos: TPoint):TColor;
                                              {return color at chosen position of interval}
      function GetColorGlobal(PercentPos: single; PatPos: TPoint):TColor;
                                              {return color at global position}
    private
      XRes, YRes: integer;                    {hor. and vert. resolution}
      ColIntList: TList;                      {list of color intervals}
    end;

implementation

const Patterns: array [0..7, 0..3, 0..3] of byte =   {Pattern definitions}
      (((1,1,1,1),(1,0,1,0),(1,1,1,1),(1,0,1,0)),   {75% gray}
       ((1,0,1,0),(0,1,0,1),(1,0,1,0),(0,1,0,1)),   {50% gray}
       ((1,0,1,0),(0,0,0,0),(1,0,1,0),(0,0,0,0)),   {25% gray}
       ((1,1,1,1),(1,0,0,0),(1,0,0,0),(1,0,0,0)),   {grid}
       ((0,1,1,0),(1,0,0,1),(1,0,0,1),(0,1,1,0)),   {squares}
       ((1,0,1,0),(1,0,1,0),(1,0,1,0),(1,0,1,0)),   {horizontal lines}
       ((1,1,1,1),(0,0,0,0),(1,1,1,1),(0,0,0,0)),   {vertical lines}
       ((1,0,0,0),(0,1,0,0),(0,0,1,0),(0,0,0,1)));  {diagnal lines}

 {* GENERAL *}

constructor TDocument.Create;
begin
  ColIntList := TList.Create;
  AddInt(0, 0, 1, clRed, clYellow, ceSolid, 1, false, isLinear, 50);
  Name := 'Default';
  Func := 'x*y';
  XF := -1;
  XT := 1;
  YF := -1;
  YT := 1;
  XRes := 300;
  YRes := 300;
  RangeF := -1;
  RangeT := 1;
  UndefColor := clBlack;
end;

destructor TDocument.Destroy;
begin
  ClearList;
  ColIntList.Free;
end;

procedure TDocument.ClearList;
var loop: integer;
begin
  for loop:=0 to ColIntList.Count-1 do
    Dispose(TPColorInterval(ColIntList.Items[loop]));     {remove list items}
  ColIntList.Clear;
end;

procedure TDocument.Copy(Source: TDocument);
var loop: integer;
begin
  Name := Source.Name;
  Func := Source.Func;
  XF := Source.XF;
  XT := Source.XT;
  YF := Source.YF;
  YT := Source.YT;
  XRes := Source.XRes;
  YRes := Source.YRes;
  RangeF := Source.RangeF;
  RangeT := Source.RangeT;
  UndefColor := Source.UndefColor;
  ClearList;
  for loop := 0 to Source.ColIntList.Count-1 do      {clone intervals}
       AddInt(loop, TPColorInterval(Source.ColIntList.Items[loop])^);
end;

procedure TDocument.SaveToStream(Stream: TStream);
var loop: integer;
begin
  Stream.WriteBuffer(Name, SizeOf(Name));
  Stream.WriteBuffer(Func, SizeOf(Func));
  Stream.WriteBuffer(XF, SizeOf(XF));
  Stream.WriteBuffer(XT, SizeOf(XT));
  Stream.WriteBuffer(YF, SizeOf(YF));
  Stream.WriteBuffer(YT, SizeOf(YT));
  Stream.WriteBuffer(RangeF, SizeOf(RangeF));
  Stream.WriteBuffer(RangeT, SizeOf(RangeT));
  Stream.WriteBuffer(XRes, SizeOf(XRes));
  Stream.WriteBuffer(YRes, SizeOf(YRes));
  Stream.WriteBuffer(UndefColor, SizeOf(UndefColor));
  Stream.WriteBuffer(ColIntList.Count, SizeOf(ColIntList.Count));
  for loop := 0 to ColIntList.Count-1 do
    with GetInt(loop)^ do begin
      Stream.WriteBuffer(FromPer, SizeOf(FromPer));
      Stream.WriteBuffer(ToPer, SizeOf(ToPer));
      Stream.WriteBuffer(LeftColor, SizeOf(LeftColor));
      Stream.WriteBuffer(RightColor, SizeOf(RightColor));
      Stream.WriteBuffer(Effect, SizeOf(Effect));
      Stream.WriteBuffer(Pattern, SizeOf(Pattern));
      Stream.WriteBuffer(Invert, SizeOf(Invert));
      Stream.WriteBuffer(IPStyle, SizeOf(IPStyle));
      Stream.WriteBuffer(LevORRand, SizeOf(LevORRand));
    end;
end;

procedure TDocument.LoadFromStream(Stream: TStream);
var loop: integer;
    NumOfIntervals: integer;
    TmpInt: TColorInterval;
begin
  ClearList;
  Stream.ReadBuffer(Name, SizeOf(Name));
  Stream.ReadBuffer(Func, SizeOf(Func));
  Stream.ReadBuffer(XF, SizeOf(XF));
  Stream.ReadBuffer(XT, SizeOf(XT));
  Stream.ReadBuffer(YF, SizeOf(YF));
  Stream.ReadBuffer(YT, SizeOf(YT));
  Stream.ReadBuffer(RangeF, SizeOf(RangeF));
  Stream.ReadBuffer(RangeT, SizeOf(RangeT));
  Stream.ReadBuffer(XRes, SizeOf(XRes));
  Stream.ReadBuffer(YRes, SizeOf(YRes));
  Stream.ReadBuffer(UndefColor, SizeOf(UndefColor));
  Stream.ReadBuffer(NumOfIntervals, SizeOf(NumOfIntervals));
  for loop := 0 to NumOfIntervals-1 do
    with TmpInt do begin
      Stream.ReadBuffer(FromPer, SizeOf(FromPer));
      Stream.ReadBuffer(ToPer, SizeOf(ToPer));
      Stream.ReadBuffer(LeftColor, SizeOf(LeftColor));
      Stream.ReadBuffer(RightColor, SizeOf(RightColor));
      Stream.ReadBuffer(Effect, SizeOf(Effect));
      Stream.ReadBuffer(Pattern, SizeOf(Pattern));
      Stream.ReadBuffer(Invert, SizeOf(Invert));
      Stream.ReadBuffer(IPStyle, SizeOf(IPStyle));
      Stream.ReadBuffer(LevORRand, SizeOf(LevORRand));
      AddInt(loop, TmpInt);
    end;
end;

 {* RESOLUTION *}

procedure TDocument.GetRes(out xr, yr: integer);
begin
  xr := XRes;
  yr := YRes;
end;

function TDocument.GetXRes:integer;
begin
  getxres := XRes;
end;

function TDocument.GetYRes:integer;
begin
  getyres := YRes;
end;

procedure TDocument.SetRes(xr, yr: integer);
begin
  if xr<=0 then xr := 1;
  if yr<=0 then yr := 1;
  XRes := xr;
  YRes := yr;
end;

 {* COLOR INTERVALS *}

function TDocument.IntCount: integer;             {return number of intervals}
begin
  intcount := ColIntList.Count;
end;

function TDocument.GetInt(index: integer):TPColorInterval;
begin
  if (index>=0) and (index<=ColIntList.Count-1) then
    getint := ColIntList.Items[index]
  else getint := nil
end;

procedure TDocument.AddInt(Index: integer; FromPer, ToPer: single; LeftColor,
                           RightColor: TColor; Effect: TColorEffect; Pattern: integer;
                           Invert: boolean; IPStyle: TIPStyle; LevORRand: single);
var PInitInt: TPColorInterval;
begin
  New(PInitInt);
  PInitInt^.FromPer := FromPer;
  PInitInt^.ToPer := ToPer;
  PInitInt^.LeftColor := LeftColor;
  PInitInt^.RightColor := RightColor;
  PInitInt^.Effect := Effect;
  PInitInt^.Pattern := Pattern;
  PInitInt^.Invert := Invert;
  PInitInt^.IPStyle := IPStyle;
  PInitInt^.LevORRand := LevORRand;
  ColIntList.Insert(Index, PInitInt);
end;

procedure TDocument.AddInt(Index: integer; Source: TColorInterval);
var PInitInt: TPColorInterval;
begin
  new(PInitInt);
  PInitInt^ := Source;
  ColIntList.Insert(Index, PInitInt);
end;

procedure TDocument.DelInt(Index: integer);
begin
  dispose(GetInt(Index));
  ColIntList.Delete(Index);
end;

procedure TDocument.DelAndResizeInt(Index: integer);
begin
  if ColIntList.Count>1 then begin
      if Index=0 then GetInt(1)^.FromPer := GetInt(0)^.FromPer
      else GetInt(Index-1)^.ToPer := GetInt(Index)^.ToPer;
      DelInt(Index);
  end;
end;

procedure TDocument.SplitInt(Index: integer);
var MidPoint: single;
begin
  with GetInt(Index)^ do begin
    MidPoint := (FromPer + ToPer) / 2;
    AddInt(Index+1, MidPoint, ToPer, LeftColor, RightColor,
           Effect, Pattern, Invert, IPStyle, LevORRand);
    ToPer := MidPoint;
  end;
end;

procedure TDocument.ChangeFromInt(Index: integer; Newrange: single);
var Index2: integer;
begin
  if (NewRange >= 0) and (Newrange < GetInt(Index)^.ToPer) then begin
    GetInt(Index)^.FromPer := NewRange;
    Index2 := Index - 1;
    while GetInt(Index2)^.FromPer > NewRange do begin
      DelInt(Index2);
      Dec(Index2);
    end;
    GetInt(Index2)^.ToPer := NewRange;
  end;
end;

procedure TDocument.ChangeToInt(Index: integer; Newrange: single);
var Index2: integer;
begin
  if (NewRange <= 1) and (Newrange > GetInt(Index)^.FromPer) then begin
    GetInt(Index)^.ToPer := NewRange;
    Index2 := Index + 1;
    while GetInt(Index2)^.ToPer < NewRange do begin
      DelInt(Index2);
    end;
    GetInt(Index2)^.FromPer := NewRange;
  end;
end;

procedure TDocument.EqualizeIntLengths;
var loop, Count: integer;
    NewRange: single;
begin
  Count := ColIntList.Count;
  for loop := 0 to Count-1 do begin
    NewRange := (loop+1) / Count;
    if loop < Count-1 then begin
      GetInt(loop)^.ToPer := NewRange;
      GetInt(loop+1)^.FromPer := NewRange;
    end;
  end;
end;

procedure TDocument.CopyIntColoring(Index: integer; Source: TColorInterval);
var Int: TPColorInterval;
begin
  Int := GetInt(Index);
  if Int <> nil then with Int^ do begin
    LeftColor := Source.LeftColor;
    RightColor := Source.RightColor;
    Effect := Source.Effect;
    Pattern := Source.Pattern;
    Invert := Source.Invert;
    IPStyle := Source.IPStyle;
    LevORRand := Source.LevORRand;
  end;
end;

function TDocument.GetColorAtInt(PercentPos: single; Int: TColorInterval; PatPos: TPoint):TColor;
        {PercentPos - percentual position in <RangeF, RangeT> interval}
        {Int - selected inteval}
        {PatPos - position in a pattern}
var TmpC: TColor;
    Par, NPar: single;
    TmpRGB: array [1..3] of integer;
    PatternBit: boolean;
begin
  Par := PercentPos;
  if PercentPos < 0 then Par := 0;
  if PercentPos > 1 then Par := 1;
  with Int do begin
    case IPStyle of
      isSquare: Par := Sqr(Par);
      isSqRoot: Par := Sqrt(Par);     {change position according to interpolation}
    end;
    NPar := 1 - Par;
    case Effect of
      ceSolid: begin
                 TmpRGB[1] := Trunc(GetRValue(LeftColor)*NPar+GetRValue(RightColor)*Par);
                 TmpRGB[2] := Trunc(GetGValue(LeftColor)*NPar+GetGValue(RightColor)*Par);
                 TmpRGB[3] := Trunc(GetBValue(LeftColor)*NPar+GetBValue(RightColor)*Par);
                 TmpC := RGB(TmpRGB[1], TmpRGB[2], TmpRGB[3]);
                 getcoloratint := TmpC;
               end;
      ceRandom: if Random < Par then getcoloratint := RightColor
                else getcoloratint := LeftColor;
      cePattern: if Patterns[Pattern, PatPos.X mod 4, PatPos.Y mod 4] = byte(Invert)
                   then getcoloratint := LeftColor
                   else getcoloratint := RightColor;
      cePatternInvert: begin
                   if Patterns[Pattern, PatPos.X mod 4, PatPos.Y mod 4] <> byte(Invert) then begin
                     NPar := Par;
                     Par := 1 - NPar;
                   end;
                   TmpRGB[1] := Trunc(GetRValue(LeftColor)*NPar+GetRValue(RightColor)*Par);
                   TmpRGB[2] := Trunc(GetGValue(LeftColor)*NPar+GetGValue(RightColor)*Par);
                   TmpRGB[3] := Trunc(GetBValue(LeftColor)*NPar+GetBValue(RightColor)*Par);
                   TmpC := RGB(TmpRGB[1], TmpRGB[2], TmpRGB[3]);
                   getcoloratint := TmpC
                 end;
      ceLeft2Pattern, cePattern2Right: begin
                   PatternBit := boolean(Patterns[pattern, PatPos.X mod 4, PatPos.Y mod 4]);
                   if (PatternBit = Invert) and (Effect = ceLeft2Pattern) then getcoloratint := LeftColor
                   else if (PatternBit <> Invert) and (Effect = cePattern2Right) then getcoloratint := RightColor
                   else begin
                     TmpRGB[1] := Trunc(GetRValue(LeftColor)*NPar+GetRValue(RightColor)*Par);
                     TmpRGB[2] := Trunc(GetGValue(LeftColor)*NPar+GetGValue(RightColor)*Par);
                     TmpRGB[3] := Trunc(GetBValue(LeftColor)*NPar+GetBValue(RightColor)*Par);
                     TmpC := RGB(TmpRGB[1], TmpRGB[2], TmpRGB[3]);
                     getcoloratint := TmpC
                   end;
                 end;
      ceLevelling: if Trunc(Par * LevORRand) mod 2 = 0
                     then getcoloratint := LeftColor
                     else getcoloratint := RightColor;
      ceLeft2Levelling, ceLevelling2Right:
                   if Trunc(Par * LevORRand) mod 2 = 0 then
                     if Effect = ceLeft2Levelling then getcoloratint := LeftColor
                     else getcoloratint := RightColor
                   else begin
                     TmpRGB[1] := Trunc(GetRValue(LeftColor)*NPar+GetRValue(RightColor)*Par);
                     TmpRGB[2] := Trunc(GetGValue(LeftColor)*NPar+GetGValue(RightColor)*Par);
                     TmpRGB[3] := Trunc(GetBValue(LeftColor)*NPar+GetBValue(RightColor)*Par);
                     TmpC := RGB(TmpRGB[1], TmpRGB[2], TmpRGB[3]);
                     getcoloratint := TmpC
                   end;
      ceRandomFill: if Random*LevORRand < 1  then getcoloratint := LeftColor
                                          else getcoloratint := RightColor;
    else getcoloratint := 0;
    end;
  end;
end;

function TDocument.GetColorGlobal(PercentPos: single; PatPos: TPoint):TColor;
var Index: integer;
    I: TColorInterval;
begin
  if PercentPos < 0 then PercentPos := 0;
  if PercentPos > 1 then PercentPos := 1;
  Index := 0;
  while GetInt(Index)^.ToPer < PercentPos do inc(Index);
  I := GetInt(Index)^;
  if I.ToPer-I.FromPer <= 0 then getcolorglobal := I.LeftColor
  else getcolorglobal :=
         GetColorAtInt((PercentPos-I.FromPer)/(I.ToPer-I.FromPer), I, PatPos);
end;

end.






