{#############################################################################}
{#                                                                           #}
{#  WinQUANT v2 - source code                                                #}
{#  Delphi version                                                           #}
{#  File: UParser.pas v2.0                                                   #}
{#                                                                           #}
{#  (C) 2001 Petr Strnsk                                                   #}
{#  All rights reserved                                                      #}
{#                                                                           #}
{#  Web:     http://www.fi.muni.cz/~xstransk/winquant                        #}
{#  E-mail:  xstransk@fi.muni.cz                                             #}
{#           aargh@email.cz                                                  #}
{#                                                                           #}
{#############################################################################}
{#                                                                           #}
{#  This file is distributed as freeware and without warranties of any kind. #}
{#  You can use it in your own applications at your own risk.                #}
{#  See help file for more information.                                      #}
{#                                                                           #}
{#############################################################################}

unit UParser;

interface

uses UParserStack, SysUtils;

type TAwaiting = (awLeftNumUnary,  {awaiting left parenthesis, number or unary operator as a next character}
                  awRightBinary);  {awaiting right parenthesis or binary operator as a next character}

{***  Parser will evaluate given function with given values of X and Y variables ***}
{***   and return a result value.                                                ***}
{***                                                                             ***}
{***      HOW TO USE TParser                                                     ***}
{***                                                                             ***}
{***  - create parser object using Create                                        ***}
{***  - REPEAT - enter a function string into ParseString                        ***}
{***           - execute Prepare                                                 ***}
{***           - if syntax error is found during analysis, exception will occur  ***}
{***           - REPEAT - enter values of X and Y variables into XVar and YVar   ***}
{***                    - execute Evaluate                                       ***}
{***                    - read result value from Result                          ***}
{***  - destroy parser object using Destroy                                      ***}

{***  Analyzed functions are stored in stack for easy manipulation               ***}
{***  Stacks contain tokens - numbers, brackets or operators                     ***}

     TParser = class
       public
         ParseString: string;  {first you must enter a function string here}
         MathError: boolean;   {was calculation completed successfully}
         Result: single;       {result value}
         XVar, YVar: single;   {values of X and Y variable}
         constructor Create;
         destructor Destroy;
         procedure Prepare;
             {after entering or changing ParserString but before using Evaluate you must execute Prepare
              to analyze the function. As long as you change only XVar and YVar afterwards, you don't
              need to execute Prepare again.}
         procedure Evaluate;
             {after changing XVar or YVar run Evaluate to get result}
       private
         Pos: integer;                  {current position in ParseString}
         PrevPos: integer;              {previous position in ParseString}

             {***  EXAMPLE:                                            ***}
             {***  ...-55+trunc(x+...                                  ***}
             {***         |    |                                       ***}
             {***   PrevPos    Pos                                     ***}

         PreparedStack: TParserStack;                           {contains analyzed function after Prepare}
         TmpStack: TParserStack;                                {working stack}
         CopyStack: TParserStack;                               {copy of PreparedStack changed by each Evaluate}
         function Read(Awaiting: TAwaiting): TStackValue;       {read token on current position}
         function Priority(Token: TStackValue): shortint;       {get priority of the operator}
         function Calculate(Token: TStackValue): TStackValue;   {get operator's value}
     end;

     EParseError = class(Exception) end;      {raised when syntax error found}

     TFnButtonTags = (btLeft, btRight {brackets},
                      btNulary, btUnary, btBinary, btMinus, btNum, btPoint, btE, btNIL);
     TPartOfNumber = (pnInteger, pnFractional, pnExponent, pnNegExp);

{***  Syntax tester is used in run-time syntax checking.                         ***}
{***  Syntax tester stores a array of token types. Before adding new token,      ***}
{***   tester checkes if token of this type should be inserted here.             ***}
{***   If not, nothing is inserted.                                              ***}
{***  For example binary operator must not be followed by right bracket. "+)" ?  ***}

     TSyntaxTester = class
       public
         constructor Create;
         destructor Destroy;
         procedure Reset;                  {clear stack and reset values}
         procedure AddTag(Tag: TFnButtonTags; Caption: string);   {try to add a token type with given caption}
         procedure DeleteTag;              {remove last inserted token}
         function GetString: string;       {get a string composed of token's captions}
       private
         LastButtonTag: TFnButtonTags;     {type of last function tag added}
         PartOfNumber: TPartOfNumber;      {which part of number are we reading now}
         OpenedBrackets: integer;          {how many opened brackets are not closed yet}
         TagArray: array of record         {aray of tokens}
                              Caption: string;
                              Tag: TFnButtonTags;
                            end;
     end;

implementation

{* PARSER *}

uses Math;

const NumberOfIdentifiers = 40;   {how many identifiers do we have}

type TIdentifiers = array [1..NumberOfIdentifiers] of record
                      N: string;
                      T: TItemType;
                    end;
const Identifiers: TIdentifiers = {longer words before shorter, because of 'SQR(T)' <-> 'SQR'}
                                  {upcase all letters}
                                  {minus symbol not included, analyzed elsewhere}
                 ((N: 'RNDSGN'; T: itRndSgn),
                  (N: 'ACOSH'; T: itfACosh), (N: 'ASINH'; T: itfASinh), (N: 'ATANH'; T: itfATanh),
                    (N: 'COTAN'; T: itfCotan), (N: 'RANDG'; T: itRandG), (N: 'ROUND'; T: itfRound),
                    (N: 'TRUNC'; T: itfTrunc),
                  (N: 'ATAN'; T: itfATan), (N: 'ACOS'; T: itfACos), (N: 'ASIN'; T: itfASin),
                    (N: 'COSH'; T: itfCosh), (N: 'FRAC'; T: itfFrac), (N: 'SINH'; T: itfSinh),
                    (N: 'SQRT'; T: itfSqrt), (N: 'TANH'; T: itfTanh),
                  (N: 'ABS'; T: itfAbs), (N: 'COS'; T: itfCos), (N: 'ANG'; T: itAng),
                    (N: 'EXP'; T: itfExp), (N: 'HPT'; T: itHypot), (N: 'LOG'; T: itLogN),
                    (N: 'MAX'; T: itMax), (N: 'MIN'; T: itMin),  (N: 'RND'; T: itRnd),
                    (N: 'SIN'; T: itfSin),(N: 'SGN'; T: itfSgn), (N: 'SQR'; T: itfSqr),
                    (N: 'TAN'; T: itfTan),
                  (N: 'LN'; T: itfLn), (N: 'PI'; T: itPi),
                  (N: 'X'; T: itXVar), (N: 'Y'; T: itYVar), (N: '+'; T: itPlus),
                    (N: '*'; T: itTimes), (N: '/'; T: itDivide), (N: '^'; T: itExp),
                    (N: '%'; T: itModulo), (N: '('; T: itLeft), (N: ')'; T: itRight));

     LeftNumUnary = [itLeft, itNum, itUnaryMinus, itfAbs, itfATan, itfCos, itfExp, itfLn,
                     itfRound, itfSin, itfSqr, itfSqrt, itfTrunc, itfACos, itfACosh,
                     itfASin, itfASinh, itfATanh, itfCosh, itfCotan, itfSinh, itfTan,
                     itfTanh, itfFrac, itfSgn, itXVar, itYVar, itRnd, itPi, itRndSgn];
                                {left bracket, number and unary operators}
     RightBinary = [itRight, itPlus, itMinus, itTimes, itDivide, itModulo, itExp,
                    itHypot, itLogN, itMax, itMin, itRandG, itAng];
                                {right bracket and binary operators}
     Operators = [itUnaryMinus, itPlus, itMinus, itTimes, itDivide, itModulo, itExp,
                  itfAbs, itfATan, itfCos, itfExp, itfLn, itfRound, itfSin, itfSqr,
                  itfSqrt, itfTrunc, itHypot, itLogN, itMax, itAng, itMin, itRandG, itfACos,
                  itfACosh, itfASin, itfASinh, itfATanh, itfCosh, itfCotan, itfSinh,
                  itfTan, itfTanh, itfFrac, itfSgn];
                                {all operators}
     Values = [itNum, itXVar, itYVar, itRnd, itPi, itRndSgn];
                                {variables, numbers and nulary operators}

constructor TParser.Create;
begin
  ParseString := '';
  MathError := false;
  XVar := 0;
  YVar := 0;
  Result := 0;
  PreparedStack := TParserStack.Create;
  TmpStack := TParserStack.Create;
  CopyStack := TParserStack.Create;
  Randomize;
end;

destructor TParser.Destroy;
begin
  PreparedStack.Destroy;
  TmpStack.Destroy;
  CopyStack.Destroy;
end;

function TParser.Read(Awaiting: TAwaiting): TStackValue;
var Check: integer;
    ParseLen: word;
    NumStr: string;
    Ch: char;
    Token: TStackValue;
    Value: single;
    loop, loop2: integer;
    Wrong: boolean;
begin
   ParseLen := Length(ParseString);
   while (Pos <= ParseLen) and (ParseString[Pos] = ' ') do Inc(Pos);
   if Pos > ParseLen then begin             {end of string}
     Token.ItemType := itEOL;
     read := Token;
     Exit;
   end;
   PrevPos := Pos;
   Ch := UpCase(ParseString[Pos]);
   if (Ch in ['0'..'9']) then begin         {read number}
     NumStr := Copy(ParseString, Pos, ParseLen-Pos+1);
     Val(NumStr, Value, Check);
     if Check <> 0 then begin
       NumStr := Copy(ParseString, Pos, Check-1);
       Val(NumStr, Value, Check);
     end;
     Token.ItemType := itNum;
     Token.Value := Value;
     Inc(Pos, Length(NumStr));
     read := Token;
     Exit;
   end

   else if Ch = '-' then begin      {unary or binary minus?}
     if Awaiting = awLeftNumUnary then Token.ItemType := itUnaryMinus
     else Token.ItemType := itMinus;
     read := Token;
     Inc(Pos);
     Exit;
   end

   else begin                       {read identifiers}
     for loop := 1 to NumberOfIdentifiers do begin
       loop2 := 1;
       Wrong := false;
       while (not Wrong) and (loop2 <= Length(Identifiers[loop].N)) do begin
         Wrong := (Pos+loop2-1 > ParseLen) or
                  (Identifiers[loop].N[loop2] <> UpCase(ParseString[Pos+loop2-1]));
         Inc(loop2);
       end;
       if not Wrong then begin
         Token.ItemType := Identifiers[loop].T;
         read := Token;
         Inc(Pos, loop2-1);
         Exit;
       end;
     end;
   end;

   Token.ItemType := itError;       {unknown identifier}
   read := Token;
   Inc(Pos);
   Exit;
end;

function TParser.Priority(Token: TStackValue): shortint;
begin     {highest priority will be resolved first}
  case Token.ItemType of
    itUnaryMinus, itfAbs, itfATan, itfCos, itfExp, itfLn, itfRound, itfSin,
    itfSqr, itfSqrt, itfTrunc, itfACos, itfACosh, itfASin, itfASinh, itfATanh,
    itfCosh, itfCotan, itfSinh, itfTan, itfTanh, itfFrac, itfSgn:
      priority := 20;
    itExp: priority := 15;
    itTimes, itDivide, itModulo: priority := 10;
    itPlus, itMinus: priority := 5;
    itHypot, itAng, itRandG, itMax, itMin, itLogN: priority := 3;
  else priority := -1;
  end;
end;

function TParser.Calculate(Token: TStackValue): TStackValue;
var Output, TmpToken: TStackValue;
    A, B: single;
begin
  Output.ItemType := itNum;
  if Token.ItemType in Operators then begin
    if TmpStack.Empty then
      raise EParseError.Create('INTERNAL ERROR: Not enough operands on position ' + IntToStr(PrevPos)) {I think this coul never happen, but who knows ... :-) }
    else TmpToken := TmpStack.Pop;       {get operands from operand stack}
    A := TmpToken.Value;
    try
      case Token.ItemType of
        itUnaryMinus: Output.Value := -A;
        itfAbs: Output.Value := Abs(A);
        itfATan: Output.Value := ArcTan(A);
        itfCos: Output.Value := Cos(A);
        itfExp: Output.Value := Exp(A);
        itfLn: if A > 0 then Output.Value := Ln(A)
               else MathError := true;
        itfRound: Output.Value := Round(A);
        itfSin: Output.Value := Sin(A);
        itfSqr: Output.Value := Sqr(A);
        itfSqrt: if A > 0 then Output.Value := Sqrt(A)
                 else MathError := true;
        itfTrunc: Output.Value := Trunc(A);
        itfACos: if (A >= -1) and (A <= 1) then Output.Value := ArcCos(A)
                 else MathError := true;
        itfACosh: Output.Value := ArcCosh(A);
        itfASin: if (A >= -1) and (A <= 1) then Output.Value := ArcSin(A)
                 else MathError := true;
        itfASinh: Output.Value := ArcSinh(A);
        itfATanh: if (A > -1) and (A < 1) then Output.Value := ArcTanh(A)
                 else MathError := true;
        itfCosh: Output.Value := Cosh(A);
        itfCotan: if Sin(A) <> 0 then Output.Value := Cotan(A)
                  else MathError := true;
        itfSinh: Output.Value := Sinh(A);
        itfTan: if Cos(A) <> 0 then Output.Value := Tan(A)
                else MathError := true;
        itfTanh: Output.Value := Tanh(A);
        itfFrac: Output.Value := Frac(A);
        itfSgn: if A > 0 then Output.Value := 1
                else Output.Value := -1;
      else
        if TmpStack.Empty then
          raise EParseError.Create('INTERNAL ERROR: Not enough operands on position ' + IntToStr(PrevPos)) {same as above}
        else TmpToken := TmpStack.Pop;
        B := TmpToken.Value;
        case Token.ItemType of
          itPlus: Output.Value := B + A;
          itMinus: Output.Value := B - A;
          itTimes: Output.Value := B * A;
          itDivide: if A <> 0 then Output.Value := B / A
                    else MathError := true;
          itModulo: if A <> 0 then Output.Value := B - Trunc(B/A)*A
                    else MathError := true;
          itExp: if ((A = Int(A)) or (B >= 0)) and ((B <> 0) or (A >= 0)) then begin
                     Output.Value := Power(B, A);
                      {That stupid Power function sometimes throws exception even if arguments
                       are in range specified by Delphi help. What the f*&# shall I do!!!!}
                 end else MathError := true;
          itHypot: Output.Value := Hypot(B, A);
          itAng: begin
                   Output.Value := ArcTan(B/A) + Pi/2;
                   if A < 0 then Output.Value := Output.Value + Pi;
                 end;
          itLogN: if (B > 1) and (A > 0) then Output.Value := LogN(B, A)
                  else MathError := true;
          itMax: Output.Value := Max(B, A);
          itMin: Output.Value := Min(B, A);
          itRandG: Output.Value := RandG(B, A);
        else Output.ItemType := itError;
        end;
      end;
      except
        on EOverFlow do raise;     {re-raise exceptions}
        on EUnderflow do raise;
        on EMathError do MathError := true;  {but not math errors}
            {It would be better to raise exception than to set MathError, but
             sometimes there are matherrors many times (try to evaluate 'ln'
             for negative numbers) and the exceptions slow the computer down
             too significantly. I wasn't able to accelerate it.}
      end;
  end
  else if Token.ItemType in Values then begin
    case Token.ItemType of    {nulary operators, numbers and variables}
      itNum: Output.Value := Token.Value;
      itXVar: Output.Value := XVar;
      itYVar: Output.Value := YVar;
      itRnd: Output.Value := Random;
      itPi: Output.Value := Pi;
      itRndSgn: if Random < 0.5 then Output.Value := -1
                else Output.Value := 1;
    end;
  end;
  calculate := Output;
end;

procedure TParser.Prepare;
var StrLen: integer;
    Awaiting: TAwaiting;
    Token: TStackValue;
begin
  PreparedStack.ClearStack;
  TmpStack.ClearStack;
  Pos := 1;
  Awaiting := awLeftNumUnary;
  StrLen := Length(ParseString);
  while (Pos <= StrLen) do begin
    Token := Read(Awaiting);

    if Token.ItemType = itError then
      raise EParseError.Create('ERROR: Syntax error on position ' + IntToStr(PrevPos));
    if ((Token.ItemType in LeftNumUnary) and (Awaiting = awRightBinary)) then
      raise EParseError.Create('ERROR: Awaiting right bracket or binary operator on position ' + IntToStr(PrevPos));
    if ((Token.ItemType in RightBinary) and (Awaiting = awLeftNumUnary)) then
      raise EParseError.Create('ERROR: Awaiting left bracket, number, variable, nulary or unary operator on position ' + IntToStr(PrevPos));

    if Token.ItemType in Values then PreparedStack.Push(Token)
    else if Token.ItemType in Operators then begin
      while (not TmpStack.Empty) and (TmpStack.Peek.ItemType <> itLeft) and
            (Priority(TmpStack.Peek) > Priority(Token)) do PreparedStack.Push(TmpStack.Pop);
      TmpStack.Push(Token);
    end
    else if Token.ItemType = itLeft then TmpStack.Push(Token)
    else if Token.ItemType = itRight then begin
      try
        while (TmpStack.Peek.ItemType <> itLeft) do PreparedStack.Push(TmpStack.Pop);
        TmpStack.Pop;
      except
        on EStackEmpty do
          raise EParseError.Create('ERROR: Too many right brackets on position ' + IntToStr(PrevPos));
      end;
    end;
    with Token do begin
      if (ItemType in Values) or (ItemType = itRight) then Awaiting := awRightBinary
      else if (ItemType in Operators) or (ItemType = itLeft) then Awaiting := awLeftNumUnary;
    end;
  end;
  if Awaiting <> awRightBinary then
    raise EParseError.Create('ERROR: Unexpected end of equation');
  while (not TmpStack.Empty) do begin
    Token := TmpStack.Pop;
    if Token.ItemType in Operators then PreparedStack.Push(Token)
    else raise EParseError.Create('ERROR: Missing right bracket on position ' + IntToStr(Pos));
  end;
  PreparedStack.Reverse;
end;

procedure TParser.Evaluate;
var Token: TStackValue;
begin
  MathError := false;
  TmpStack.ClearStack;
  CopyStack.Copy(PreparedStack);
  while not CopyStack.Empty do begin
    Token := CopyStack.Pop;
    if Token.ItemType in Values then TmpStack.Push(Calculate(Token))
    else if Token.ItemType in Operators then TmpStack.Push(Calculate(Token))
    else raise EParseError.Create('INTERNAL ERROR: Operators error');
    if MathError then Exit;
  end;
  if TmpStack.Empty then raise EParseError.Create('ERROR: Unexpected end of equation');
  Result := TmpStack.Pop.Value;
  if not TmpStack.Empty then raise EParseError.Create('ERROR: Unexpected end of equation');
end;

{* SYNTAX TESTER *}

constructor TSyntaxTester.Create;
begin
  Reset;
end;

destructor TSyntaxTester.Destroy;
begin
  Reset;
end;

procedure TSyntaxTester.Reset;
begin
  LastButtonTag := btNIL;
  OpenedBrackets := 0;
  PartOfNumber := pnInteger;
  SetLength(TagArray, 0)
end;

procedure TSyntaxTester.AddTag(Tag: TFnButtonTags; Caption: string);
begin
   case LastButtonTag of      {tag not allowed?}
     btNIL, btLeft, btBinary, btUnary:
       if not(Tag in [btLeft, btNum, btNulary, btMinus, btUnary]) then Exit;
     btMinus:
       if (PartOfNumber = pnNegExp) and (Tag <> btNum) then Exit
       else if (PartOfNumber <> pnNegExp) and not(Tag in [btLeft, btNum, btNulary, btMinus, btUnary]) then Exit;
     btRight:
       if not(Tag in [btRight, btBinary, btMinus]) then Exit;
     btNulary:
       if not(Tag in [btRight, btBinary, btMinus]) then Exit;
     btNum: begin
              if not(Tag in [btRight, btBinary, btMinus, btNum, btPoint, btE]) then Exit;
              if ((PartOfNumber = pnFractional) and (Tag = btPoint)) or
                 ((PartOfNumber = pnExponent) and (Tag in [btPoint, btE])) then Exit;
            end;
     btPoint:
       if Tag <> btNum then Exit;
     btE:
       if not(Tag in [btNum, btMinus]) then Exit;
   end;
   if (Tag = btRight) and (OpenedBrackets = 0) then Exit;  {too many right brackets}
   case Tag of      {which part of number are we reading}
     btPoint: PartOfNumber := pnFractional;
     btE: PartOfNumber := pnExponent;
     btMinus: if LastButtonTag = btE then PartOfNumber := pnNegExp
              else PartOfNumber := pnInteger
   else
     if Tag <> btNum then PartOfNumber := pnInteger;
   end;
   case Tag of
     btLeft: Inc(OpenedBrackets);
     btRight: Dec(OpenedBrackets);
   end;
   LastButtontag := Tag;

   Caption := LowerCase(Caption);            {add tag}
   SetLength(TagArray, Length(TagArray)+1);
   case Tag of
     btUnary: begin
                Caption := Caption + '(';    {unary operator automatically opens a bracket}
                Tag := btLeft;
                Inc(OpenedBrackets);
              end;
     btBinary: Caption := ' ' + Caption + ' ';
   end;
   TagArray[High(TagArray)].Caption := Caption;
   TagArray[High(TagArray)].Tag := Tag;
end;

procedure TSyntaxTester.DeleteTag;
begin
   if Length(TagArray) > 0 then begin
     case TagArray[High(TagArray)].Tag of
       btLeft: Dec(OpenedBrackets);
       btRight: Inc(OpenedBrackets);
     end;
     SetLength(TagArray, Length(TagArray)-1);
     if Length(TagArray) = 0 then LastButtonTag := btNIL
     else LastButtonTag := TagArray[High(TagArray)].Tag;
  end;
end;

function TSyntaxTester.GetString: string;
var Tmp: string;
    loop: integer;
begin
  Tmp := '';
  for loop := 0 to High(TagArray) do Tmp := Tmp + TagArray[loop].Caption;
  getstring := Tmp;
end;

end.
